###############################################################################
# Compute the 4 by 4 spectral density matrix estimate for the given vector-
# valued series
###############################################################################

est.spectral.matrix <- function( x1, y1, x2, y2, K=10 )
{
	N <- length(x1)

	d1.x <- fft(x1 - mean(x1))
	d1.y <- fft(y1 - mean(y1))
	d2.x <- fft(x2 - mean(x2))
	d2.y <- fft(y2 - mean(y2))
	
	x <- rbind(d1.x, d1.y, d2.x, d2.y) / sqrt(2*pi*N)

	n <- N%/%(2*K)-1
	j <- seq(2, by=K, length=n)
	f <- array(complex(4*4*n), c(4, 4, n))
	for (i in 1:n)
	{
		f[,,i] <- (as.matrix(x[,j[i]:(j[i]+K-1)]) %*% 
			t(Conj(as.matrix(x[,j[i]:(j[i]+K-1)])))) / K
	}
	f
}

###############################################################################
# Compute the 4 by 4 spectral density matrix estimate for the given drifters.
# The drifters can be either two real drifters or two simulated drifters.
###############################################################################

spec.matrix <- function( drifter1, drifter2, K=10 )
{
	if (charmatch(names(drifter1)[1], "days",nomatch=0))
	{
		a <- max( min(drifter1$days), min(drifter2$days) )
		b <- min( max(drifter1$days), max(drifter2$days) )
		ind1 <- drifter1$days >= a & drifter1$days <= b
		ind2 <- drifter2$days >= a & drifter2$days <= b
	}
	else
	{
		n <- min(length(drifter1$lat), length(drifter2$lat))
		ind1 <- 1:n
		ind2 <- 1:n
	}

	x1 <- diff(drifter1$lat[ind1])
	y1 <- diff(drifter1$long[ind1])
	x2 <- diff(drifter2$lat[ind2])
	y2 <- diff(drifter2$long[ind2])

	est.spectral.matrix( x1, y1, x2, y2, K=K )
}

###############################################################################
# Plot the modulus of the 4 by 4 spectral density matrix estimate at all 
# calculated frequencies with the same scale in each panel.  
###############################################################################

plot.spec <- function( f )
{
	x <- numeric(0)		# Put the spectra on the same scale
	for (i in 1:4)
	  for (j in 1:4)
	    x <- c(x, Mod(f[i,j,]))

	par(mfrow=c(4,4))
	for (i in 1:4)
	  for (j in 1:4)
	    tsplot(Mod(f[i,j,]), ylim=range(x))
	par(mfrow=c(1,1))
	invisible()
}

###############################################################################
# Compute the (inverse) square root of a matrix
###############################################################################

sqrt.matrix <- function( mat, inverse=F )
{
	e <- eigen(mat)

	e$val[ e$val < 0 & abs(e$val) < 1e-06 ] <- 0

# The above line is equivalent to the following: 
#	if (sum(e$val < 0))
#          for (i in 1:length(e$val))
#	    if (e$val[i] < 0 & abs(e$val[i]) < 1e-06) e$val[i] <- 0

	if (inverse)
	  as.matrix(e$vec) %*% diag(as.complex(1/sqrt(e$val))) %*%
		t(Conj(as.matrix(e$vec)))
	else
	  as.matrix(e$vec) %*% diag(as.complex(sqrt(e$val))) %*%
		t(Conj(as.matrix(e$vec)))
}

###############################################################################
# Compute the coherency from the spectral density estimate at all calculated
# frequencies.  
###############################################################################

compute.coh <- function( f )
{
	n <- dim(f)[3]	
	coh <- array(complex(2*2*n), c(2,2,n))
	eigen.val <- matrix(complex(2*n), ncol=2)
	for (i in 1:n)
  	{
		sqrt.11 <- sqrt.matrix(f[1:2,1:2,i], T)
		sqrt.22 <- sqrt.matrix(f[3:4,3:4,i], T)
		coh[,,i] <- sqrt.11 %*% f[1:2,3:4,i] %*% sqrt.22
	}
	coh	
}

###############################################################################
# Plot the modulus, real part, imaginary part, or argument of the
# coherency according to the flag.
###############################################################################

plot.coh <- function( coh, modulus=T, real=F, imag=F, arg=F )
{
	par(mfrow=c(2,2))
	if (modulus)
	{
		tsplot(Mod(coh[1,1,]), ylim=c(-1,1))
		tsplot(Mod(coh[1,2,]), ylim=c(-1,1))
		tsplot(Mod(coh[2,1,]), ylim=c(-1,1))
		tsplot(Mod(coh[2,2,]), ylim=c(-1,1))
	}
	else if (real)
	{
		tsplot(Re(coh[1,1,]), ylim=c(-1,1))
		tsplot(Re(coh[1,2,]), ylim=c(-1,1))
		tsplot(Re(coh[2,1,]), ylim=c(-1,1))
		tsplot(Re(coh[2,2,]), ylim=c(-1,1))
	}
	else if (imag)
	{
		tsplot(Im(coh[1,1,]), ylim=c(-1,1))
		tsplot(Im(coh[1,2,]), ylim=c(-1,1))
		tsplot(Im(coh[2,1,]), ylim=c(-1,1))
		tsplot(Im(coh[2,2,]), ylim=c(-1,1))
	}
	else if (arg)
	{
		tsplot(Arg(coh[1,1,]), ylim=c(-1,1))
		tsplot(Arg(coh[1,2,]), ylim=c(-1,1))
		tsplot(Arg(coh[2,1,]), ylim=c(-1,1))
		tsplot(Arg(coh[2,2,]), ylim=c(-1,1))
	}
	par(mfrow=c(1,1))
	invisible()
}

###############################################################################
# Simulate N by 4 data with the given spectral density matrix (in some sense)
###############################################################################

sim.dep <- function( f, N=500 )
{
	# Compute coherency at each frequency
	coh <- compute.coh( f )

	# Average the coherencies
	for (i in 2:n)
		coh[,,1] <- coh[,,1] + coh[,,i]
	coh <- coh[,,1] / n
#cat("average coherency\n")
#coh <- matrix(complex(real=c(1,0,0,1)), ncol=2) # force coh = I
#print(coh)

	# Form the constant spectral density matrix
	f.hat <- diag(as.complex(rep(1,4)))
	f.hat[1:2,3:4] <- coh
	f.hat[3:4,1:2] <- t(Conj(coh))
#cat("f.hat\n")
#print(f.hat)

	# Compute the sqrt of f.hat
	sqrt.f.hat <- sqrt.matrix(f.hat)
#cat("sqrt.f.hat\n")
#print(sqrt.f.hat)

	# Generate N complex normals (0, I)
	z <- matrix(complex(real=rnorm(4*N, 0, 1/sqrt(2)),
			im=rnorm(4*N, 0, 1/sqrt(2))), nrow=4)
	z[,1] <- complex(real=Re(z[,1]), im=0)
	z[,(N/2)+1] <- complex(real=Re(z[,(N/2)+1]), im=0)
	z[,((N/2)+2):N] <- Conj(z[,rev(2:(N/2))])
	 
	for (i in 1:4)
		z[i,] <- fft(z[i,], inverse=T) / sqrt(N)

	t(Re(sqrt.f.hat %*% z))
}

###############################################################################
# Compute the "trace" statistic of the spectral density matrix
###############################################################################

trace.stat <- function( f )
{
#	N <- dim(f)[3]
#	tr <- complex(N)
#	for (i in 1:N)
#		tr[i] <- sum(diag(f[1:2,1:2,i] %*% solve(f[3:4,3:4,i])))
#	Re(tr)
	Re(apply(f,3,function(x) sum(diag(x[1:2,1:2] %*% solve(x[3:4,3:4])))))
}

###############################################################################
# Compute the "determinant" statistic of spectral density matrix
###############################################################################

det.stat <- function( f )
{
#	N <- dim(f)[3]
#	d <- complex(N)
#	for (i in 1:N)
#		d[i] <- det(f[1:2,1:2,i] %*% solve(f[3:4,3:4,i]))
#	Re(d)
	Re(apply(f,3,function(x) det(x[1:2,1:2] %*% solve(x[3:4,3:4]))))
}

##############################################################################
# Test if either the trace or the determinant sample computed from f 
# has the same distribution as the corresponding sample generated from
# Monte Carlo simulation.  It returns the p-value of the KS two-sample
# test.
##############################################################################

indep.spec.test <- function( f, K=10, method="determinant" )
{
	imethod <- charmatch(method, c("trace", "determinant"),
		nomatch=0)
	switch(imethod+1,
		stop("method is unknown"),
		p <- kstwo(trace.stat(f), wishart.trace[[K]]),
		p <- kstwo(det.stat(f), wishart.det[[K]])
	)
	p
}

###############################################################################
# Test if either the trace or the determinant sample computed from f
# has the same distribution as the corresponding sample computed from g.
# It returns the p-value of the KS two-sample test.
###############################################################################

dep.spec.test <- function( f, g, method="determinant" )
{
	imethod <- charmatch(method, c("trace", "determinant"),
		nomatch=0)
	switch(imethod+1,
		stop("method is unknown"),
		p <- kstwo(trace.stat(f), trace.stat(g)),
		p <- kstwo(det.stat(f), det.stat(g)),
		)
	p
}

###############################################################################
# Compute N trace statistics or determinant statistics of (W1 *
# W2^(-1)) from independent normal sample with K ranges from 2
# to 10.  
#
# Note:  When K=1, the spectral density matrix is unsmoothed, so
# singularity will occur when computing the statistics.
###############################################################################

wishart.sample <- function( N, method="determinant" )
{
	imethod <- charmatch(method, c("trace", "determinant"),
		nomatch=0)

	K <- 1:10
	n <- (N+1) * (2*K)
	w <- vector("list", 10)

	switch(imethod+1, 
	stop("method is unknown"),
	{
		for (i in 2:10)	
		{
		  f <- est.spectral.matrix(rnorm(n[i]), rnorm(n[i]),
			rnorm(n[i]), rnorm(n[i]), K[i])
		  w[[i]] <- trace.stat(f)
		}
	}
	,
	{
		for (i in 2:10)		
		{
		  f <- est.spectral.matrix(rnorm(n[i]), rnorm(n[i]),
			rnorm(n[i]), rnorm(n[i]), K[i])
		  w[[i]] <- det.stat(f)
		}
	}
	)
	w
}





